/*
 *
 *	This is free software. You can redistribute it and/or modify under
 *	the terms of the GNU Library General Public License version 2.
 * 
 *	Copyright (C) 1999 by kra
 *
 */
#ifndef __QP_THR_H
#define __QP_THR_H

/*
 * 
 * Threads
 * 
 */

class QpRunnable: public QpBase {
    private:
	QpClonedException *r_except;
    public:
	void InvokeMain();
	
	QpRunnable(const char *name = NULL);
	virtual ~QpRunnable();
	virtual void Raise();

	virtual void Main() = 0;
};

typedef void (*signal_func)(int sig_nr, void *arg);
struct reg_sig {
	int 	     r_sig;
	signal_func  r_func;
	void	     *r_func_arg;
	QpRunnable  *r_job;
	unsigned int r_magic;
};

class QpThread: public QpBase {
    protected:
	pthread_t t_id;

	QpMutex	t_mutex;
	bool    t_was_joined;
	QpCond	t_join_cond;
	
	bool	t_cancel;
	
	bool	t_sleep;
	QpCond	t_sleep_cond;
	bool	t_init;

	unsigned int t_prio;
	unsigned int t_stack_size;
	void 	     *t_stack_addr;
	QpMutex	     t_start_mutex;
	
	QpClonedException *t_except;
	QpRunnable  	*t_runnable;

	list<reg_sig *> s_sync_list;
	
	static bool 		t_first;
	static pthread_t 	t_main_id;
	static QpThread		*t_main_thread;
	static pthread_key_t 	t_curr_key;

	static QpMonitor	t_count_monitor;
	static int		t_count;
	static sigset_t		t_orig_sigmask;
	
	static void *invoke_main(void *arg);
    protected:
	virtual void Main();	/* The Main function */
	
	static void WaitForAllThreads();
	virtual void ThrInit(unsigned int prio = QP_PRIO_DFL, unsigned int stack_size = 0);
	void __Begin(void *stack_addr, bool set_tid = true, bool set_sigmask = true);
	void __Finish();
	void __OrigSigmask();
    protected:
	QpThread(unsigned int prio = QP_PRIO_DFL,
		 unsigned int stack_size = 0, const char *name = NULL,
		 bool thr_created = false);
    public:
	QpThread(QpRunnable *r, unsigned int prio = QP_PRIO_DFL,
		unsigned int stack_size = 0, const char *name = NULL);
	virtual ~QpThread();
	
	void Start();
	bool Join(unsigned int msec, bool throw_ex = true);
	bool Join(const struct timespec *ts = NULL, bool throw_ex = true);
	bool JoinAbs(const struct timespec *ts = NULL, bool throw_ex = true);

#if 0
	static void Terminate() {
		Current()->__Terminate();
		pthread_exit(NULL);
	}
#endif
	static bool Sleep(unsigned int msec);
	static bool Sleep(const struct timespec *ts = NULL);
	static bool SleepAbs(const struct timespec *ts = NULL);
	
	void WakeUp();
	
	static void Yield() { sched_yield();}
	
	void Cancel();
	static void TestCancel();
	
	void Raise();

	bool IsAlive();

	
	unsigned int SetPriority(unsigned int prio);
	unsigned int GetPriority();
	
	static QpThread *MainThread();
	static QpThread *Current();
	static unsigned long CurrentSeq();
	unsigned long ThreadSeq();
	
    friend class QpInit;
    friend class QpSignal;
    private:
	QpThread(const QpThread&);
	QpThread& operator = (const QpThread&);
};


class QpThreadPool;

class QpPThread: public QpThread {
    private:
	QpThreadPool   *p_pool;
	unsigned int	p_life_time;
	QpRunnable     *p_cur_job;
    public:
	QpPThread(QpThreadPool *pool, unsigned int prio = QP_PRIO_DFL,
		  unsigned int stack_size = 0,
		  unsigned int msec_life_time = 0,
		  const char *name = NULL);
	virtual ~QpPThread() { Join();}
	virtual void Main();
    friend class QpThreadPool;
    private:
	QpPThread(const QpPThread&);
	QpPThread& operator = (const QpPThread&);
};

class QpThreadPool: public QpBase {
    protected:
	QpMutex		p_mutex;
	bool		p_running;
	
	list<QpRunnable *> p_job;
	QpCond 	   	   p_job_cond;
	
	list<QpRunnable *> p_stat;
	QpCond		   p_stat_cond;

	list<QpPThread *>   p_threads;
	vector<QpPThread *> p_threads_clean;
	QpCond		    p_clean_cond;
	
	unsigned int 	  p_prio;
	unsigned int 	  p_stack_size;
	unsigned int 	  p_max_threads;
	unsigned int 	  p_life_time;
    private:
	void ThreadsClean();
    public:
	QpThreadPool(unsigned int prio = QP_PRIO_DFL,
		  unsigned int stack_size = 0,
		  unsigned int max_threads = 1, 
		  bool on_demand = false,
		  unsigned int life_time = 0,
		  const char *name = NULL);
	~QpThreadPool();
	
	void Do(QpRunnable *j);
	QpRunnable *Wait(QpRunnable *what, unsigned int msec, bool throw_ex = true);
	QpRunnable *Wait(QpRunnable *what = NULL, const struct timespec *ts_rel = NULL, bool throw_ex = true);
	QpRunnable *WaitAbs(QpRunnable *what = NULL, const struct timespec *ts_abs = NULL, bool throw_ex = true);
	
	int CurrentThreads();
	QpRunnable *CurrentJob(unsigned int thr_nr = 0);
    friend class QpPThread;
    private:
	QpThreadPool(const QpThreadPool&);
	QpThreadPool& operator = (const QpThreadPool&);
};

#endif
